#!/usr/bin/env python
import sys, re, os, pickle, copy, csv
from optparse import OptionParser
from scipy.stats import ttest_ind
from random import normalvariate 
from math import ceil

class DictObject(object):
  def __getitem__(self, key):
    return self.dict[key]
  def keys(self):
    return self.dict.keys()
  def values(self):
    return self.dict.values()
  def pop(self, key):
    return self.dict.pop(key)
  def count(self):
    return len(self.keys())
  def copy(self):
   return copy.deepcopy(self)
  def get(self, attribute, value):
    dictObject = self.copy()
    for key in dictObject.dict.keys():
      try:
        keyValue = getattr(dictObject.dict[key], attribute)()
      except TypeError:
        keyValue = getattr(dictObject.dict[key], attribute)
      except AttributeError: # mslFound not there
        keyValue = None
      if not keyValue == value:
        dictObject.dict.pop(key)
    return dictObject
  def index(self, attribute):
    index = {}
    for key in self.dict.keys():
      index[key] = getattr(self.dict[key], attribute)
    return index
  def sortedList(self, attribute):
    def sortByValue(dict): # replaced with builtin sorted() in Python2.4
      items = [(v, k) for k, v in dict.items()]
      items.sort()
      items.reverse()
      items = [(k, v) for v, k in items]
      return items
    sortedList = []
    for keyValueTuple in sortByValue(self.index(attribute)):
      sortedList.append(self.dict[keyValueTuple[0]])
    return sortedList
  def indexByMethod(self, method):
    index = {}
    for key in self.dict.keys():
      index[key] = eval('self.dict[key].'+method+'()')
    return index
  def sortedListByMethod(self, method):
    def sortByValue(dict): # replaced with builtin sorted() in Python2.4
      items = [(v, k) for k, v in dict.items()]
      items.sort()
      items.reverse()
      items = [(k, v) for v, k in items]
      return items
    sortedList = []
    for keyValueTuple in sortByValue(self.indexByMethod(method)):
      sortedList.append(self.dict[keyValueTuple[0]])
    return sortedList

  
class Gemoda(object):
  def __init__(self, peakLibraryLibrary,options=None):
    self.peakLibraryLibrary = peakLibraryLibrary
    if options:
      self.supportThreshold(options.supportThreshold.strip().lower())
    else:
      self.supportThreshold('medium')
    self.i = peakLibraryLibrary.getFastaFileString() # Fasta sequence file
    self.l = 1 # word size
    if options:
      self.similarityThreshold(options.similarityThreshold.strip().lower())
    else:
      self.similarityThreshold('medium')
    self.c = 0 # cluster method (0: clique-finding)
    self.m = 2 # similarity metric (2: AMDIS match factor)
    self.noSelfCompare = 0
    if options:
      self.elutionThreshold(options.elutionThreshold.strip().lower())
    else:
      self.elutionThreshold('medium')
    if peakLibraryLibrary.count() == 1: # single sample mode
      self.singlePeakLibraryMode()
  def singlePeakLibraryMode(self):
    #  We could use self.supportThreshold('repeated'), but we'll leave it as = 2
    #   just in case someday we change what 'repeated' is.  And I'll leave this 
    #   comment in so that when I look for 'repeated' or supportThreshold, I'll
    #   appropriately find it here.
    self.k = 2 # support
    self.c = 1 # similarity metric (1: single-linkage clustering)
    self.e *= 2 # double elution to bridge gap
    self.p = 1
  def elutionThreshold(self, arg): # max allowable elution difference
    elutionThresholdDict = {\
        'high' : 0.5,
        'medium' : 1,
        'low' : 2,
        'off' : 1E5
        }
    self.e = elutionThresholdDict[arg]
  def supportThreshold(self, arg): # min support required
    max = self.peakLibraryLibrary.count()
    supportThresholdDict = {\
        'high' : max,
        'medium' : int(round(0.75*max)),
        'low' : int(round(0.50*max)),
        'repeated' : 2
        }
    if supportThresholdDict[arg] > 2:
      self.p = supportThresholdDict[arg]
      self.k = supportThresholdDict[arg]
    else:
      self.p = 2
      self.k = 2
  def similarityThreshold(self, arg): # max allowable elution difference
    similarityThresholdDict = {\
        'high' : 0.9,
        'medium' : 0.8,
        'low' : 0.7,
        'verylow' : 0.6
        }
    self.g = similarityThresholdDict[arg]
  def getOutputFileString(self):
    return self.i + '.go'
  def run(self):
    self.peakLibraryLibrary.outputFasta()
    # OK, here's a simpler way than looking in the path... we'll first try to
    #  use the src directory beneath us to execute gemoda... but if gemoda-r
    #  doesn't exist there, we'll default back to using it from the main
    #  installation in the path.  Otherwise, we quit and cry.
    scriptPath = os.path.split(sys.argv[0])[0]
    tryPath = os.path.join(os.path.join(scriptPath,'src'),'gemoda-r') 
    if os.path.exists(tryPath):
      command = tryPath
    else:
      command = 'gemoda-r'
    self.argList = [\
        command,
        '-p', str(self.p),
        '-i', str(self.i), # must not contain spaces
        '-l', str(self.l),
        '-k', str(self.k),
        '-g', str(self.g),
        '-c', str(self.c),
        '-m', str(self.m),
        '-e', str(self.e),
        '-n',
        '-j',
        '-s', '0',
        '-o', str(self.getOutputFileString())
        ]
    if (self.noSelfCompare == 1):
      self.argList.append('-u')
    outputF.write(' '.join(self.argList)+'\n')
    outputF.flush()
    # Note that this error-checking is currently not working, because it comes
    #  back as a success even if gemoda cannot be found.
    pid = os.spawnv(os.P_NOWAIT,command,self.argList)
    (pid,code) = os.waitpid(pid,0)
    if (code > 0):
      os.remove(self.i)
      outputF.write("Error running Gemoda!  Exit code: "+str(code)+"\n")
      outputF.flush()
      sys.exit("Error running Gemoda!  Exit code: "+str(code)+"\n")
    os.remove(self.i)
  def getPatternLibrary(self):
    self.run()
    
    isInputFile = re.compile(
        '^Input\ file\ =\ (?P<inputFile>.*)')
    isGemodaParam = re.compile(
        '^l = (?P<l>[0-9]*), k = (?P<k>[0-9]*), g = (?P<g>[0-9|.]*)')
    isMinSeq = re.compile(
        '^Minimum # of sequences with motif = (?P<minSeq>[0-9]*)')
    isRecursivePruning = re.compile(
        '^Recursive pruning is (?P<recursivePruning>[A-Z]*)')
    isOamSize = re.compile(
        '^oam->size = (?P<oamSize>[0-9]*)')
    isPattern = re.compile(
        '^pattern (?P<id>[0-9]*):\tlen=(?P<len>[0-9]*)\tsup=(?P<sup>[0-9]*)')
    isSequencePosition = re.compile(
        '^[ ]*(?P<sequence>[0-9]*)\t(?P<position>[0-9]*)\t(?P<centroidMatch>[0-9|.]*)')

    patternLibrary = PatternLibrary(self.peakLibraryLibrary)
    patLibFile = open(self.getOutputFileString())
    for line in patLibFile:
      if isInputFile.match(line):
        inputFile = isInputFile.findall(line)[0]
        setattr(patternLibrary, 'inputFile', inputFile)
      elif isGemodaParam.match(line):
        (l, k, g) = isGemodaParam.findall(line)[0]
        setattr(patternLibrary, 'l', l)
        setattr(patternLibrary, 'k', k)
        setattr(patternLibrary, 'g', g)
      elif isMinSeq.match(line):
        minSeq = isMinSeq.findall(line)[0]
        setattr(patternLibrary, 'minSeq', minSeq)
      elif isRecursivePruning.match(line):
        recursivePruning = isRecursivePruning.findall(line)[0]
        setattr(patternLibrary, 'recursivePruning', recursivePruning)
      elif isOamSize.match(line):
        oamSize = isOamSize.findall(line)[0]
        setattr(patternLibrary, 'oamSize', oamSize)
      elif isPattern.match(line):
        (id, _len, sup) = isPattern.findall(line)[0]
        pattern = Pattern(int(id), int(_len), int(sup))
      elif isSequencePosition.match(line):
        (sequence, position, centroidMatch) = \
            isSequencePosition.findall(line)[0]
        # We need to copy the peak here
        # Otherwise, the pattern contains the last read peak at all entries
        # I'm not actually sure why
        # MPS: I think I may know why...
        #  ... because Python likes to pass complicated
        #  things by reference rather than by value.  So, if one were to
        #  just assign peak to the desired item without copying, Python
        #  quick-copies and uses a pointer to peak, which will later change
        #  in value (it in turn points to some other real peak object).  
        #  This may be circumvented by not using the assignment to "peak"
        #  as an intervening variable... I think.  But this works, so it's 
        #  fine.
        peak = self.peakLibraryLibrary[int(sequence)][int(position)].copy()
        peak.centroidMatch = float(centroidMatch)
        pattern.add(peak)
        if pattern.count() == pattern.sup:
          patternLibrary.add(pattern)
    patLibFile.close()
    os.remove(self.getOutputFileString())
    return patternLibrary

class PatternLibrary(DictObject):
  def __init__(self, peakLibraryLibrary):
    self.peakLibraryLibrary = peakLibraryLibrary
    self.counter = 0
    self.dict = {}
  def add(self, pattern):
    self.dict[self.counter] = pattern
    self.counter += 1
  def renumber(self):
    self.counter = 0
    for key in self.keys():
      self.add(self.dict.pop(key))
  def getPeakLibrary(self):
    peakLibrary = PeakLibrary(self.peakLibraryLibrary.conditionPath)
    #peakLibrary.descriptor = 'PatternLibrary'
    peakLibrary.descriptor = self.peakLibraryLibrary.getCondition()
    peakLibrary.minRA = self.peakLibraryLibrary.minRA
    peakLibrary.sampSize = self.peakLibraryLibrary.sampSize
    for pattern in self.values():
      peakRef = None
      RAlist = []
      ISlist = []
      AMlist = []
      RTlist = []
      sourceList = []
      for peak in pattern.values():
        if (not peakRef) and (peak.centroidMatch == 1):
#          peakLibrary.add(peak)
          peakRef = peak
        RAlist.append(peak.RA)
        try:
          ISlist.append(peak.IS)
        except: 
          pass
        try:
          AMlist.append(peak.AM)
        except:
          pass
        RTlist.append(peak.RT)
        # Let's only bring along the name and subdirectory
        sourceList.append(\
          os.path.join(os.path.basename(os.path.dirname(peak.source)),\
          os.path.basename(peak.source)))
      peakRef.RAlist = RAlist
      try:
        peakRef.ISlist = ISlist
      except:
        pass
      try:
        peakRef.AMlist = AMlist
      except:
        pass
      peakRef.RTlist = RTlist
      peakRef.sourceList = sourceList
      peakRef.RA = sum(peakRef.RAlist)/float(len(peakRef.RAlist))
      if peakLibrary.sampSize > len(RAlist):
        peakRef.RApval = copy.deepcopy(peakRef.RAlist)
        # Should we use makeZeroMeanSamps here?  I'm torn.  Before, we used
        #  it to massage our desired mean and standard deviation into
        #  the t-test.  Now, we don't care about the mean and standard
        #  deviation, and we shouldn't feel too bad about just setting it 
        #  to zero, so we'll do that for now.
        #peakRef.RApval.extend(makeZeroMeanSamps(peakLibrary.minRA,\
          #peakLibrary.minRA, pattern.sup - len(RAlist))
        for i in range(peakLibrary.sampSize - len(RAlist)):
          peakRef.RApval.append(0)
      else:
        peakRef.RApval = peakRef.RAlist
      peakLibrary.add(peakRef)
    return peakLibrary
  def gemodaOutput(self, file=sys.stdout):
    for pattern in self.values():
      file.write('pattern ' + str(pattern.id)\
          + ':\tsup=' + str(pattern.count()) + '\n')
      for peak in pattern.values():
        file.write('\t' + str(peak.sequence) + '\t' + str(peak.position)\
            + '\t' + '%1.3f'%peak.centroidMatch + '\t\n')
      file.write('\n')
  def output(self, file=sys.stdout):
    for pattern in self.values():
      file.write('pattern ' + str(pattern.id)\
          + ':\tsup=' + str(pattern.count()) + '\n')
      for peak in pattern.values():
        file.write('\tRT:' + '%1.1f'%peak.RT + '\tName:' + peak.name\
            + '\tMatch:' + '%1.3f'%peak.centroidMatch)
        try:
          file.write('\tAmt:' + str(peak.size()) + '%')
        except:
          pass
        file.write('\t\n')
  def removeOverlaps(self, patternKey, derefArray):
    for peak in self.dict[patternKey].values():
      seqNum = peak.sequence
      posNum = peak.position
      # Each pattern named removalKey will be removed from the pattern list,
      #  and all references to that pattern must also be removed.
      for removalKey in derefArray[seqNum][posNum]:
        if removalKey == patternKey:
          continue
        else:
          # Note that what follows will also remove the removalKey from
          #  the node that stays in the kept pattern.
          for otherPeak in self.dict[removalKey].dict.values():
            otherSeq = otherPeak.sequence
            otherPos = otherPeak.position
            if len(derefArray[otherSeq][otherPos]) > 1:
              derefArray[otherSeq][otherPos].remove(removalKey)
            else:
              derefArray[otherSeq][otherPos] = [-1]
          self.dict[removalKey].removed = True




class Pattern(DictObject):
  def __init__(self, id, len, sup):
    self.id = id
    self.len = len
    self.sup = sup
    self.dict = {}
    self.counter = 0
    self.removed = False
  def add(self, peak):
    self.dict[self.counter] = peak
    self.counter += 1

class PeakLibraryLibrary(DictObject):
  def __init__(self, conditionPath):
    self.counter = 0
    self.dict = {}
    self.conditionPath = conditionPath
    self.minRA = 1000
  def add(self, peakLibrary):
    peakLibrary.setSequence(self.counter)
    self.dict[self.counter] = peakLibrary
    self.counter += 1
  def getCondition(self):
    return os.path.split(self.conditionPath)[1]
  def getFastaFileString(self):
    if self.count() > 1:
      return os.path.join(
          self.conditionPath, self.getCondition() + '.fa')
    else:
      [peakLibrary] = self.values()
      return os.path.join(self.conditionPath, peakLibrary.descriptor + '.fa')
  def outputFasta(self):
  # We want to know what the max Mz is so that we don't leave out any of
  #  of the peaks.
    maxMz = self.getPllMaxMz()
    fastaFile = open(self.getFastaFileString(), 'w')
    for sample in self.values():
      sample.outputFasta(fastaFile,maxMz)
    fastaFile.close()
  def eliminateInternalMatches(self,options=None):
    for sample in self.values():
      sample.eliminateInternalMatches(options)
  def define(self): 
    def getEluPathList():
      eluPathList = []
      for fileString in os.listdir(self.conditionPath):
        if fileString.endswith('.ELU') or fileString.endswith('.elu'):
          eluPathList.append(os.path.join(self.conditionPath, fileString))
      eluPathList.sort()
      return eluPathList
    eluPathList = getEluPathList()
    for eluPath in eluPathList:
      #eluFilePath = eluPath
      peakLibrary = PeakLibrary(self.conditionPath)
      peakLibrary.define(eluPath, open(eluPath))
      if peakLibrary.minRA < self.minRA:
        self.minRA = peakLibrary.minRA
      self.add(peakLibrary)
    self.sampSize = len(eluPathList)
  # Contol function to loop over the inidividual peak libraries.
  def getPllMaxMz(self):
    max = 0
    for sample in self.values():
      temp = sample.getPlMaxMz()
      if (temp > max):
        max = temp
    return max


# this is also like an msl
class PeakLibrary(DictObject):
  def __init__(self, conditionPath):
    self.counter = 0
    self.dict = {}
    self.conditionPath = conditionPath
    self.minRA = 1000
  def add(self, peak):
    if not hasattr(peak, 'alreadysetname'):
      peak.name += '_' + str(peak.getPMaxIntenMz()) + 'mz'
      peak.name += '_' + str(peak.RT) + 'min'
      peak.alreadysetname = True
    peak.position = self.counter
    self.dict[self.counter] = peak
    self.counter += 1
  def renumber(self):
    self.counter = 0
    for key in self.keys():
      self.add(self.dict.pop(key))
  def setSequence(self, sequence):
    self.sequence = sequence
    for peak in self.values():
      peak.sequence = sequence
  def setDescriptor(self):
    eluString = os.path.split(self.eluPath)[1]
    self.descriptor = os.path.splitext(eluString)[0]
  def define(self, eluPath, file): # ELU or MSL file
    self.eluPath = eluPath
    self.setDescriptor()
    isName = re.compile('^[Nn][Aa][Mm][Ee]:(?P<name>.*)')
    isForm = re.compile('[Ff][Oo][Rr][Mm]:(?P<form>.*)')
    isCasNo = re.compile('[Cc][Aa][Ss][Nn][Oo]:(?P<casNo>.*)')
    isRi = re.compile('[Rr][Ii]:(?P<ri>.*)')
    isRw = re.compile('[Rr][Ww]:(?P<rw>.*)')
    isRt = re.compile('[Rr][Tt]:(?P<rt>.*)')
    isSource = re.compile('[Ss][Oo][Uu][Rr][Cc][Ee]:(?P<source>.*)')
    isRe = re.compile('[Rr][E]')
    isNumPeaks = re.compile('^[Nn][Uu][Mm] [Pp][Ee][Aa][Kk][Ss]:[ ]*(?P<numPeaks>[0-9]{1,3})')
    isBlank = re.compile('^$')
    isFloat = re.compile('^[0-9]*\.[0-9]*$')
    isInt = re.compile('^[0-9]$')
    # this nameFlag allows us to see if we've gotten the name of the 
    #  spectrum yet... if not, we won't let "blanks" allow for submission
    #  of peaks.  This hedges against multiple lines in between entries
    #  and blank lines at the beginning of the file.
    location = None
    nameFlag = 0
    for line in file:
      line = line.strip()
      # set new relative location if necessary
      if isName.match(line):
        peak = Peak()
        if eluPath.endswith('.ELU') or eluPath.endswith('.elu'):
          peak.source = eluPath
        location = 'name'
        nameString = isName.findall(line)[0]
        nameFlag = 1
        if nameString.count('|'):
          peak.name = self.descriptor + '_' + str(self.counter)
          for attributeValuePair in nameString.split('|'):
            if attributeValuePair.count(':'):
              [attribute, value] = attributeValuePair.split(':')
            else:
              attribute = attributeValuePair[0:2]
              value = attributeValuePair[2:]
            if isFloat.match(value):
              value = float(value)
            elif isInt.match(value):
              value = int(value)
            setattr(peak, attribute, value)
        else:
          peak.name = nameString
      elif isForm.match(line):
        location = 'form'
        form = isForm.findall(line)[0]
        peak.form = form
      elif isCasNo.match(line):
        location = 'casNo'
        casNo = isCasNo.findall(line)[0]
        peak.casNo = casNo
      elif isRi.match(line):
        location = 'ri'
        ri = isRi.findall(line)[0]
        peak.ri = ri
      elif isRw.match(line):
        location = 'rw'
        rw = isRw.findall(line)[0]
        peak.rw = rw
      elif isRt.match(line):
        location = 'rt'
        rt = isRt.findall(line)[0]
        #peak.rt = rt
        peak.RT = float(rt)
      elif isSource.match(line):
        location = 'source'
        source = isSource.findall(line)[0]
        peak.source = source
      elif isRe.match(line):
        location = 're'
      elif isNumPeaks.match(line):
        location = 'numPeaks'
        numPeaks = int(isNumPeaks.findall(line)[0])
        peak.numPeaks = numPeaks
        # We don't want to search the line if it is the NumPeaks line,
        #  so we implement a flag that tells us that it's the NumPeaks
        #  line, and subsequent 'numPeaks' locations will work.
        skipFlag = 1
      elif isBlank.match(line):
        location = 'blank' 
        if nameFlag == 1:
          if eluPath.endswith('.ELU') or eluPath.endswith('.elu'):
            if peak.RA > minPeakHeight:
              if peak.RA < self.minRA:
                self.minRA = peak.RA
              self.add(peak)
          else:
            self.add(peak)
          nameFlag = 0
      else:
        pass
      # read mass spectrum when applicable
      if location == 'numPeaks':
      # skipFlag is used to catch files not ending with blank lines
        if skipFlag == 1:
          skipFlag = 0
        else:
          peak.read(line)
    # Must catch case where file does not end with blank line... means
    #  it was left in location = 'numPeaks'.
    if location == 'numPeaks':
      if skipFlag == 1:
        outputF.write('Incomplete input file.')
        outputF.flush()
        raise StandardError('Incomplete input file.')
      else:
        if eluPath.endswith('.ELU') or eluPath.endswith('.elu'):
          if peak.RA > minPeakHeight:
            if peak.RA < self.minRA:
              self.minRA = peak.RA
            self.add(peak)
        else:
          if peak.RA < self.minRA:
            self.minRA = peak.RA
          self.add(peak)  
      
    outputF.write('READ: ' + str(self.count()) + ' peaks from ' \
        + file.name + '\n')
    outputF.flush()
        #+ eluPath + '\n')
  def outputFasta(self, fastaFile, maxMz):
    fastaFile.write('> ' + self.descriptor + '\n')
    for peak in self.values():
      fastaFile.write(str(peak.RT) + ' ')
    fastaFile.write('\n')
    for mz in range(0, maxMz+1):
      for peak in self.values():
        fastaFile.write('%0.1f'%peak.getIntensity(mz) + ' ')
      fastaFile.write('\n')
  def outputMsl(self, mslFile):
    try:
      maxPvalue = .01*float(2)/(len(self.values()) * self.numClasses * \
        (self.numClasses-1))
    except:
      pass
    for peak in self.values():
      mslFile.write('NAME:' + peak.name + '\n')
      mslFile.write('CASNO:' + '' + '\n')
      mslFile.write('RI:' + ', '*10 + '\n')
      mslFile.write('RW:' + ', '*10 + '\n')
      mslFile.write('RT: ' + '%1.3f'%float(peak.RT) + '\n')
      try:
        mslFile.write('COMMENTS:' + 'sizeRank=' + str(peak.sizeRank) + '/' \
          + str(self.count()) + '\n')
      except AttributeError: # no sizeRank
        pass
      try:
        # Here we add in delimiters for classes in which the metabolite
        #  does not occur.
        peak.classExist.sort()
        maxClass = peak.classExist[-1]
        for i in range(maxClass):
          if peak.classExist[i] != i+1:
            peak.classExist.insert(i,'-')
        extender = []
        for i in range(len(peak.classExist),self.numClasses):
          extender.append('-')
        peak.classExist.extend(extender)
        # Get rid of them ugly single quotes in the output!
        mslFile.write('COMMENTS: In conditions: ' + \
          str(peak.classExist).replace("'",'') + '\n')
      except:
        pass
      try:
        str(peak.pvalList)
#        mslFile.write('COMMENTS: p-value list: [class, class, t-test p-value]\n')
        commentFlag = False
        for tuple in peak.pvalList:
          if tuple[2] < maxPvalue:
            if not commentFlag:
              mslFile.write('COMMENTS: [condition, condition, p-value] for significant t-tests\n')
              commentFlag = True
            mslFile.write('COMMENTS: [' + str(tuple[0]) + ', ' + str(tuple[1]) + ', ' + '%1.3le'%tuple[2] + ']\n') 
      except:
        pass
      try:
        mslFile.write('COMMENTS: RA ' + str(peak.RA) + '\n')
      except:
        pass
#      try:
#        mslFile.write('COMMENTS: ' + str(peak.RAdict.values()) + '\n')
#      except:
#        pass
      mslFile.write('SOURCE:' + peak.source + '\n')
      mslFile.write('NUM PEAKS:  ' + str(peak.count()) + ' \n')
      peak.outputMslIntensities(mslFile)
      mslFile.write('\n')
  def outputDetailsFile(self,detailsFile):
    try:
      maxPvalue = .01*float(2)/(len(self.values()) * self.numClasses * \
        (self.numClasses-1))
    except:
      pass
    for peak in self.values():
      detailsFile.write('NAME:' + peak.name + '\n')
      detailsFile.write('Centroid RT: ' + '%1.3f'%float(peak.RT) + '\n')
      detailsFile.write('Source\tRT\tRA\tIS\tAM\n')
      try:
      # If we don't have these totList's, the first step here will throw
      #  an error.  totList's imply multiple classes
        for i in range(len(peak.RTtotList)):
          for j in range(len(peak.RTtotList[i])):
            detailsFile.write(peak.sourceTotList[i][j] + '\t')
            detailsFile.write(str(peak.RTtotList[i][j]) + '\t')
            detailsFile.write(str(peak.RAtotList[i][j]) + '\t')
            try:
              detailsFile.write(str(peak.IStotList[i][j]) + '\t')
            except:
              pass
            try:
              detailsFile.write(str(peak.AMtotList[i][j]) + '\n')
            except:
              pass
        commentFlag = False
        for tuple in peak.pvalList:
          if tuple[2] < maxPvalue:
            if not commentFlag:
              detailsFile.write('Differential: [condition, condition, p-value]\n')
              commentFlag = True
            detailsFile.write('Differential: [' + str(tuple[0]) + ', ' + \
              str(tuple[1]) + ', ' + '%1.3le'%tuple[2] + ']\n')        
      except:
      # So we don't have lists... just execute this for one condition then.
        for i in range(len(peak.RTlist)):
          detailsFile.write(peak.sourceList[i] + '\t')
          detailsFile.write(str(peak.RTlist[i]) + '\t')
          detailsFile.write(str(peak.RAlist[i]) + '\t')
          try:
            detailsFile.write(str(peak.ISlist[i]) + '\t')
          except:
            pass
          try:
            detailsFile.write(str(peak.AMlist[i]) + '\n')
          except:
            pass
      detailsFile.write('\n')
      
  def eliminateInternalMatches(self,options=None):
    def sortByValue(dict):
      items = [(v, k) for k, v in dict.items()]
      items.sort()
      items.reverse()
      items = [(k, v) for v, k in items]
      return items
      
    peakLibraryLibrary = PeakLibraryLibrary(self.conditionPath)
    peakLibraryLibrary.add(self)
    
    gemoda = Gemoda(peakLibraryLibrary,options)
    patternLibrary = gemoda.getPatternLibrary()
    counter = 0
    for pattern in patternLibrary.values():
      myPatternDict = {}
      for peakIndex in pattern.keys():
        peak = pattern[peakIndex]
        myPatternDict[peakIndex] = peak.size()
      items = sortByValue(myPatternDict)
      for item in items[1:]:
        peak = pattern[item[0]]
        self.dict.pop(peak.position)
        counter += 1
    self.renumber()
    outputF.write('Eliminated ' + str(counter) \
        + ' internally matching peaks.\n')
    outputF.flush()
  def outputMatrixDict(self, file, eluAttribute):
    writer = csv.writer(file, dialect='excel')
    assert self.eluDescriptors
    matrixDict = self.getMatrixDict(eluAttribute)
    writer.writerow([''] + self.eluDescriptors)
    for mslPeak in self.values():
      row = [mslPeak.name]
      for eluDescriptor in self.eluDescriptors:
        try:
          value = matrixDict[mslPeak.name][eluDescriptor]
        except:
          value = 0.
        row.append(value)
      writer.writerow(row)
  def getMatrixDict(self, eluAttribute):
    matrixDict = {}
    for mslPeak in self.values():
      assert self.eluDescriptors
      for eluDescriptor in self.eluDescriptors:
        if hasattr(mslPeak, 'matchDict'):
          if mslPeak.matchDict.has_key(eluDescriptor):
            eluPeak = mslPeak.matchDict[eluDescriptor]
            value = getattr(eluPeak, eluAttribute)
            if matrixDict.has_key(mslPeak.name):
              matrixDict[mslPeak.name][eluDescriptor] = value
            else:
              matrixDict[mslPeak.name] = {eluDescriptor : value}
          else:
            pass
        else:
          pass
    return matrixDict
  def compareToElu(self, eluPeakLibrary, options):
    # archive the descriptor in a list
    try:
      self.eluDescriptors.append(eluPeakLibrary.descriptor)
    except:
      self.eluDescriptors = [eluPeakLibrary.descriptor]

    # get a pattern library
    peakLibraryLibrary = PeakLibraryLibrary(self.conditionPath)
    peakLibraryLibrary.add(self)
    peakLibraryLibrary.add(eluPeakLibrary)
    gemoda = Gemoda(peakLibraryLibrary)
    if not options.matrixRT: # automatically replace with better match
      gemoda.elutionThreshold('off')
    else:
      # use the gemoda class elution threshold library
      gemoda.elutionThreshold(options.elutionThreshold)
      requiredRT = gemoda.e
    gemoda.similarityThreshold(options.similarityThreshold)
    if options.libraryThreshold:
      gemoda.similarityThreshold(options.libraryThreshold)
    gemoda.noSelfCompare = 1
    patternLibrary = gemoda.getPatternLibrary()

    # find the highest matching mslPeak for each eluPeak
    libHitDict = {'mslPosition' : {}, 'match' : {}}
    def defineHit(libHitDict, eluPosition, mslPosition, matchFactor):
      if libHitDict['mslPosition'].has_key(eluPosition):
          libHitDict['mslPosition'][eluPosition] = mslPosition
          libHitDict['match'][eluPosition] = matchFactor
      else:
        libHitDict['mslPosition'][eluPosition] = mslPosition
        libHitDict['match'][eluPosition] = matchFactor

    for pattern in patternLibrary.values():
      mslPeakFromPattern = pattern.get('sequence', 0)[0]
      mslPosition = mslPeakFromPattern.position
      mslPeak = self.dict[mslPosition]
      eluPeakFromPattern = \
          pattern.get('sequence', 1).sortedList('centroidMatch')[0]
      eluPosition = eluPeakFromPattern.position
      eluPeak =  eluPeakLibrary.dict[eluPosition]
      matchFactor = compare(mslPeak, eluPeak)

      
      # CHANGES by MPS, 6/1/2007: removed RT enforcement here, put it
      #  above, which keeps us from performing useless calculations in
      #  Gemoda that can be costly for very big libraries.
      # for each elu peak, set the best msl by match factor
      #  possibly subject to a retention time criteria
      if libHitDict['mslPosition'].has_key(eluPosition): # previously hit
        if libHitDict['match'][eluPosition] < matchFactor: # ie, better match
          #if not options.matrixRT: # automatically replace with better match
            defineHit(libHitDict, eluPosition, mslPosition, matchFactor)
          #else: # enforce RT criteria!
          #  if abs(eluPeak.RT - mslPeak.RT) < requiredRT: # come back
          #    defineHit(libHitDict, eluPosition, mslPosition, matchFactor)
          #  else: # not close enough RT!
          #    pass
        else: # do not overwrite a better matching peak
          pass
      else: # new hit
        #if not options.matrixRT: # automatically define a new hit
          defineHit(libHitDict, eluPosition, mslPosition, matchFactor)
        #else: # enforce RT criteria!
        #  if abs(eluPeak.RT - mslPeak.RT) < requiredRT: # come back
        #    defineHit(libHitDict, eluPosition, mslPosition, matchFactor)
        #  else: # not close enough RT!
        #    pass
      
      # for each elu peak, set the best msl by match factor
      #if libHitDict['mslPosition'].has_key(eluPosition):
        #if libHitDict['match'][eluPosition] < matchFactor:
          # overwrite old hit with better library match
         # defineHit(libHitDict, eluPosition, mslPosition, matchFactor)
          
          #libHitDict['mslPosition'][eluPosition] = mslPosition
          #libHitDict['match'][eluPosition] = matchFactor
        #elif libHitDict['match'][eluPosition] == matchFactor:
          # when elu and library peak consist of one ion and match 100%
        #  oldMslPosition = libHitDict['mslPosition'][eluPosition]
        #  oldMslPeak = self.dict[oldMslPosition]
        #  if abs(eluPeak.RT - mslPeak.RT) < abs(eluPeak.RT - oldMslPeak.RT):
        #    # overwrite old hit with equally matching closer library match
        #    libHitDict['mslPosition'][eluPosition] = mslPosition
        #    libHitDict['match'][eluPosition] = matchFactor
        #  else:
        #    pass # ie, do not overwrite equally matching but more distant match 
       # else:
       #   pass # ie, do not overwrite if not a better library match
      #else: # define new hit
      #  libHitDict['mslPosition'][eluPosition] = mslPeakFromPattern.position
      #  libHitDict['match'][eluPosition] = matchFactor
    
    # for each msl peak, set the best elu peak by size
    for eluPosition in libHitDict['mslPosition'].keys():
      mslPosition = libHitDict['mslPosition'][eluPosition]
      mslPeak = self.dict[mslPosition]
      eluPeak = eluPeakLibrary.dict[eluPosition]
      if hasattr(mslPeak, 'matchDict'):
        if mslPeak.matchDict.has_key(eluPeakLibrary.descriptor):
          currentEluPeak = mslPeak.matchDict[eluPeakLibrary.descriptor]
          if currentEluPeak.size() < eluPeak.size():
            # if bigger, replace
            mslPeak.matchDict[eluPeakLibrary.descriptor] = eluPeak
          else:
            pass # ie, do _not_ overwrite a smaller peak
        else:
          mslPeak.matchDict[eluPeakLibrary.descriptor] = eluPeak
      else:
        mslPeak.matchDict = {eluPeakLibrary.descriptor : eluPeak}
  def compareToMsl(self, mslPeakLibrary, options):
    peakLibraryLibrary = PeakLibraryLibrary(self.conditionPath)
    peakLibraryLibrary.add(self)
    peakLibraryLibrary.add(mslPeakLibrary)
    gemoda = Gemoda(peakLibraryLibrary)
    gemoda.elutionThreshold('off')
    gemoda.similarityThreshold(options.similarityThreshold)
    if options.libraryThreshold:
      gemoda.similarityThreshold(options.libraryThreshold)
    gemoda.noSelfCompare = 1
    patternLibrary = gemoda.getPatternLibrary()
    isRetentionName = re.compile('^ *[0-9|.]* min (?P<name>.*)')
    alreadyMatched = re.compile('(.+?)\s*\(Match=([\d\.]+)%\)')
    idCounter = 0
    for pattern in patternLibrary.values():
      patternPeak = pattern.get('sequence', 0)[0]
      mslPeak = pattern.get('sequence', 1).sortedList('centroidMatch')[0]
      # Right here we can see over-assignment... if we have a match to the
      #  same peak, then whichever one is done last will be the final
      #  result in the msl.
      peak = self.dict[patternPeak.position]
      peak.mslFound = 1
      mslName = mslPeak.name
      # So we implement a check to make sure that we haven't already matched
      #  it... if we have, then we check to see which one is better.
      if not alreadyMatched.search(peak.name):
        idCounter += 1
        if isRetentionName.match(mslName):
          peak.name = isRetentionName.findall(mslName)[0]
        else:
          peak.name = mslName
      # OK, this is worth noting... what we have done here is to use the
      #  fact that our compiled peaks have been checked against themselves
      #  first, so there will only be one match from "sequence 0".  Then, there
      #  will be a list of matches from the library, as some may be self-
      #  similar.  Above we took the best-scoring match.  Now, if the pattern
      #  we found in sequence 0 was the centroid, the best scoring match
      #  is less than one, and that will be right.  If the centroid is in
      #  the list of library entries, then we will have selected that one,
      #  and so the distance to that centroid is the centroid match of our
      #  pattern peak.
        peak.name += ' (Match=' \
          + '%1.0f'%(100*min(patternPeak.centroidMatch, mslPeak.centroidMatch)) + '%)'
      else:
        alreadyMatchedObject = alreadyMatched.search(peak.name)
        matchedName = alreadyMatchedObject.group(1)
        matchedFactor = int(alreadyMatchedObject.group(2))
#        print matchedFactor, 100*patternPeak.centroidMatch, 100*mslPeak.centroidMatch,100*min(patternPeak.centroidMatch,mslPeak.centroidMatch)
#        print (matchedFactor < 100*min(patternPeak.centroidMatch,mslPeak.centroidMatch))
        if (matchedFactor < 100*min(patternPeak.centroidMatch,mslPeak.centroidMatch)):
#          print "reassigning " + matchedName + " to "
          if isRetentionName.match(mslName):
            peak.name = isRetentionName.findall(mslName)[0]
          else:
            peak.name = mslName
#          print peak.name
          peak.name += ' (Match=' \
            + '%1.0f'%(100*min(patternPeak.centroidMatch, mslPeak.centroidMatch)) + '%)'
    outputF.write(str(idCounter)\
        + ' identifications made from ' + str(self.count()) + ' total peaks.\n')
    outputF.flush()
  def assignSizeRank(self):
    sizeRank = 1
    for peak in self.sortedListByMethod('size'):
      peak.sizeRank = sizeRank
      sizeRank += 1
  # This function gets the max Mz value from a peak library... is called
  #  by getPllMaxMz.
  def getPlMaxMz(self):
    max = 0
    for peak in self.values():
      temp = peak.getPMaxMz()
      if (temp > max):
        max = temp
    return max

class Peak(DictObject):
  def __init__(self):
    self.dict = {}
    self.RT = 0.0
    self.RAlist = []
  def size(self):
    return self.RA
  def getIntensity(self, mz):
    try:
      return self.dict[mz]
    except KeyError:
      return 0
  def read(self, line):
    isMz = re.compile(\
      '\s*\(\s*(?P<mz>[0-9]{1,4})[,| ]*(?P<intensity>[0-9]{1,4})[ ]*(?P<notes>.*?)\)')
    isAltMz = re.compile(\
      '\s*(?P<mz>[0-9]{1,4})[,| ]+(?P<intensity>[0-9]{1,4})(?P<notes>.*?);')
    if isMz.match(line):
      for (mz, intensity, notes) in isMz.findall(line):
        if notes == '':
          self.dict[int(mz)] = int(intensity)
    elif isAltMz.match(line):
      for (mz, intensity, notes) in isAltMz.findall(line):
        if notes == '':
          self.dict[int(mz)] = int(intensity)
    else:
      outputF.write('MSL peak format not supported!')
      outputF.flush()
      raise StandardError('MSL peak format not supported!') 
  def outputMslIntensities(self, mslFile):
    counter = 0
    mzList = self.keys()
    mzList.sort()
    for mz in mzList:
      (row, col) = divmod(counter, 5)
      intensity = self.dict[mz]
      mslFile.write('(' + '%3.0f'%mz + ' ' + '%4.0f'%intensity + ') ')
      counter += 1
      if col == 4 or counter == self.count():
        mslFile.write('\n')
  # Here we actually get the maximum Mz for a given peak.  This is called
  #  by get PlMaxMz.  Note that in Python, the last element in array can
  #  be accessed by array[-1]
  def getPMaxMz(self):
    mzList = self.keys()
    mzList.sort()
    return mzList[-1]
  def getPMaxIntenMz(self):
    intenList = self.values()
    intenList.sort()
    maxInten = intenList[-1]
    return self.keys()[self.values().index(maxInten)]
      

# allLibs is a peakLibraryLibrary
def compareAllLibraries(allLibs, basePath, options):
  # In keeping with current convention, we name this gemoda
  gemoda = Gemoda(allLibs, options)
  gemoda.supportThreshold('repeated')
  gemoda.noSelfCompare = 1
  masterPatternLibrary = gemoda.getPatternLibrary()
  #  OK, we want to record where each sequence,position pair (or peak)
  #   occurs across all of the patterns so we easily know which patterns
  #   must be added for our agglomeration.
  derefArray = []
  for library in allLibs.values():
    tempArray = []
    for peak in library.values():
      tempArray.append([])
    derefArray.append(tempArray)
  for key, pattern in masterPatternLibrary.dict.items():
    for peak in pattern.values():
      derefArray[peak.sequence][peak.position].append(key)
  dictList = masterPatternLibrary.dict.items()
  i = 0
  while i < len(dictList):
    key = dictList[i][0]
    pattern = dictList[i][1]
    # OK, we're going to go through and make sure that all appropriate
    #  pruning or combining is done.  Start with the first pattern...
    #  if it needs agglomerating, then do it and prune.  If not, continue.
    if pattern.removed == True:
      i += 1
      continue
    keyList = []
    toBeAdded = [key]
    while(toBeAdded):
      keyList.append(toBeAdded[0])
      addedPattern = masterPatternLibrary.dict[toBeAdded[0]]
      for peak in addedPattern.values():
        for clique in derefArray[peak.sequence][peak.position]:
          if (clique not in keyList) and (clique not in toBeAdded):
            toBeAdded.append(clique)
      toBeAdded.pop(0)
    if (len(keyList) == 1):
        i += 1
        continue
    # OK, now we have a list of cliques that are all connected somehow
    # First we need to figure which is the biggest
    sizeList = []
    for index in keyList:
      sizeList.append(len(masterPatternLibrary.dict[index].values()))
    maxSize = max(sizeList)
    if sizeList.count(maxSize) > 1:
      candidateSizeListIndexes = []
      for i in range(len(sizeList)):
        if sizeList[i] == maxSize:
          candidateSizeListIndexes.append(i)
      candidateKeys = []
      for i in range(len(candidateSizeListIndexes)):
        candidateKeys.append(keyList[candidateSizeListIndexes[i]])
      candidatePatterns = []
      for i in range(len(candidateKeys)):
        candidatePatterns.append(masterPatternLibrary.dict[candidateKeys[i]])
      maxPatternIndex = chooseBest(candidatePatterns)
      maxPatternKey = candidateKeys[maxPatternIndex]
      masterPatternLibrary.removeOverlaps(maxPatternKey,derefArray)
    else:
      maxPatternKey = keyList[sizeList.index(maxSize)]
      masterPatternLibrary.removeOverlaps(maxPatternKey,derefArray)
  #Now we parse the derefArray to make our final superLibrary.
  addedPatternIndexes = []
  superLibrary = PeakLibrary(basePath)
  superLibrary.descriptor = os.path.split(os.path.abspath(basePath))[1]
  listOfLibIndex = range(len(allLibs.values()))
  comboList = generatePathCombinations(listOfLibIndex)
  sampSizeList = []
  for library in allLibs.values():
    sampSizeList.append(library.sampSize)
  averageSampSize = int(ceil(sum(sampSizeList)/float(len(sampSizeList))))
  nullRaList = makeZeroMeanSamps(allLibs.minRA, allLibs.minRA, averageSampSize)
  for i in range(len(allLibs.values())):
    for j in range(len(allLibs.values()[i].values())):
      if derefArray[i][j] == [-1]:
        pass
      else:
        # Here we note that an empty set evaluates to false.
        if derefArray[i][j]:
          RAdict = {}
          RAtotList = []
          IStotList = []
          AMtotList = []
          RTtotList = []
          sourceTotList = []
          classExist = []
          peakRef = None
          # Note that now, no peak is in more than one clique, so either the
          #  number was -1 (so it was eliminated), it's empty (so it had no
          #  match), or it has one single number in it.
          for peak in masterPatternLibrary.dict[derefArray[i][j][0]].values():
            if (not peakRef) and (peak.centroidMatch == 1):
              peakRef = peak
            RAdict[peak.sequence] = peak.RApval
            RAtotList.append(peak.RAlist)
            try:
              IStotList.append(peak.ISlist)
            except:
              pass
            try:
              AMtotList.append(peak.AMlist)
            except:
              pass
            RTtotList.append(peak.RTlist)
            sourceTotList.append(peak.sourceList)
            classExist.append(peak.sequence + 1)
            derefArray[peak.sequence][peak.position] = [-1]
          peakRef.pvalList = []
          peakRef.classExist = copy.deepcopy(classExist)
          for k in range(len(allLibs.values())):
            if k+1 not in classExist:
              RAdict[k] = nullRaList
          for combo in comboList:
            pval = ttest_ind(RAdict[combo[0]],RAdict[combo[1]])
            peakRef.pvalList.append([combo[0]+1,combo[1]+1,pval[1]])
          peakRef.RAdict = copy.deepcopy(RAdict)
          peakRef.RAtotList = copy.deepcopy(RAtotList)
          try:
            peakRef.IStotList = copy.deepcopy(IStotList)
          except:
            pass
          try:
            peakRef.AMtotList = copy.deepcopy(AMtotList)
          except:
            pass
          peakRef.RTtotList = copy.deepcopy(RTtotList)
          peakRef.sourceTotList = copy.deepcopy(sourceTotList)
          delattr(peakRef,'sizeRank')
          delattr(peakRef,'RA')
          superLibrary.add(peakRef)
        else:
          peak = allLibs.dict[i].dict[j]
          RAdict = {}
          RAtotList = [peak.RAlist]
          try:
            IStotList = [peak.ISlist]
          except:
            pass
          try:
            AMtotList = [peak.AMlist]
          except:
            pass
          RTtotList = [peak.RTlist]
          sourceTotList = [peak.sourceList]
          peak.classExist = [peak.sequence + 1]
          RAdict[peak.sequence] = peak.RApval
          for k in range(len(allLibs.values())):
            if k+1 not in peak.classExist:
              RAdict[k] = nullRaList
          peak.pvalList = []
          for combo in comboList:
            pval = ttest_ind(RAdict[combo[0]],RAdict[combo[1]])
            peak.pvalList.append([combo[0]+1,combo[1]+1,pval[1]])
          peak.RAdict = copy.deepcopy(RAdict)
          peak.RAtotList = copy.deepcopy(RAtotList)
          try:
            peak.IStotList = copy.deepcopy(IStotList)
          except:
            pass
          try:
            peak.AMtotList = copy.deepcopy(AMtotList)
          except:
            pass
          peak.RTtotList = copy.deepcopy(RTtotList)
          peak.sourceTotList = copy.deepcopy(sourceTotList)
          delattr(peak,'sizeRank')
          delattr(peak,'RA')
          superLibrary.add(peak)
  superLibrary.numClasses = len(allLibs.values())
  return superLibrary

def makeZeroMeanSamps(minRA, stdDev, numSamps):
  nullRaList = []
  for i in range(numSamps):
    nullRaList.append(normalvariate(minRA,stdDev))
  return nullRaList


def chooseBest(listOfPatterns):
  i = 0
  bestIndex = 0
  bestScore = 10000
  for pattern in listOfPatterns:
    try:
      sum = pattern.simScore
    except:
      sum = 0
      comboList = generatePathCombinations(pattern.keys())
      for combo in comboList:
        score = compare(pattern.dict[combo[0]],pattern.dict[combo[1]])
        sum += pow(1 - score, 2)
      pattern.simScore = sum
    if sum < bestScore:
      bestScore = sum
      bestIndex = i
    i += 1
  return bestIndex
    
def compare(peakOne,peakTwo):
  sum1 = 0
  sum2 = 0
  sum3 = 0

  mzList = []
  mzList.extend(peakOne.keys())
  mzList.extend(peakTwo.keys())
  mzList.sort()
  rangeList = range(len(mzList)-1)
  rangeList.reverse()
  for i in rangeList:
    if mzList[i] == mzList[i-1]:
      mzList.pop(i)
  for mz in mzList:
    sum1 += pow(peakOne.getIntensity(mz) * peakTwo.getIntensity(mz), .5) * \
      pow(mz, 2)
    sum2 += peakOne.getIntensity(mz) * pow(mz,2)
    sum3 += peakTwo.getIntensity(mz) * pow(mz,2)
  return float(pow(sum1, 2)) / (sum2 * sum3)

def getMslPeakLibrary(options):
  def getMslFileString():
    for fileString in os.listdir(options.basePath):
      if fileString.endswith('.MSL') or fileString.endswith('.msl'):
        if not fileString.startswith('spectConnect'):
          outputF.write('Using ' + fileString + ' library.\n')
          outputF.flush()
          return fileString
    raise StandardError('MSL library cannot be found.')
  mslFileString = getMslFileString()
  mslFilePath = os.path.join(options.basePath, mslFileString)
  mslPeakLibrary = PeakLibrary(options)
  mslFile = open(mslFilePath)
  mslPeakLibrary.define(mslFileString, mslFile)
  mslFile.close()
  return mslPeakLibrary


def main():
  parser=OptionParser()
  parser.add_option('', '--basePath', dest='basePath',
      default='ELU2')
  parser.add_option('', '--outputFile', dest='outputFile',
      default=None)
  parser.add_option('', '--minPeakHeight', dest='minPeakHeight',
      default='0')
  parser.add_option('', '--supportThreshold', dest='supportThreshold',
      default='medium')
  parser.add_option('', '--similarityThreshold', dest='similarityThreshold',
      default='medium')
  parser.add_option('', '--elutionThreshold', dest='elutionThreshold',
      default='medium')
  parser.add_option('', '--keepTemps', dest='keepTemps', action="store_true",
      default=False)
  parser.add_option('', '--libraryThreshold', dest='libraryThreshold',
      default=False)
  parser.add_option('', '--onlyMatrix', dest='onlyMatrix',
      action='store_true', default=False)
  parser.add_option('', '--alsoMatrix', dest='alsoMatrix',
      action='store_true', default=False)
  parser.add_option('', '--matrixDataType', dest='matrixDataType',
      default='RA')
  parser.add_option('', '--matrixDoNotEnforceRT', dest='matrixDoNotEnforceRT',
      action='store_true', default=False)
  parser.add_option('', '--matrixEnforceRT', dest='matrixEnforceRT',
      action='store_true', default=False)
  (options, args) = parser.parse_args()

  # set default behavior to use RT if alsoMatrix
  # and not to use RT if only matrix
  # (unless overridden of course)
  if options.onlyMatrix:
    if not options.matrixEnforceRT:
      setattr(options, 'matrixRT', False)
    else:
      setattr(options, 'matrixRT', True)
  elif options.alsoMatrix:
    if not options.matrixDoNotEnforceRT:
      setattr(options, 'matrixRT', True)
    else:
      setattr(options, 'matrixRT', False)

  global minPeakHeight
  libraryFlag = 1
  basePath = options.basePath
  minPeakHeight = float(options.minPeakHeight)
  global outputF
  if options.outputFile:
    try:
      outputF = open(options.outputFile,'w')
    except:
      sys.stdout.write("Unable to make output file... printing to stdout.")
      outputF = sys.stdout
      outputF.flush()
  else:
    outputF = sys.stdout
  conditionPathList = getConditionPathList(basePath)
  for path in conditionPathList:
    if path.endswith('spectStorage'):
      conditionPathList.pop(conditionPathList.index(path))
  resultsPath = os.path.join(basePath,'results')
  try:
    os.mkdir(resultsPath)
  except:
    outputF.write("\n\nCan't create results directory; it may already " \
      + "exist.\n  If so, save whatever results you want and delete the " \
      + "directory.\n")
    outputF.flush()
    raise StandardError("\n\nCan't create results directory; it may already " \
      + "exist.\n  If so, save whatever results you want and delete the " \
      + "directory.\n")
  try:
    mslPeakLibrary = getMslPeakLibrary(options)
  except:
    outputF.write('MSL cannot be found.\n')
    outputF.flush()
    libraryFlag = 0
  totalLib = PeakLibraryLibrary(os.path.abspath(basePath))
  if options.onlyMatrix:
    assert mslPeakLibrary
    generateMatrix(mslPeakLibrary, options)
  else:
    for conditionPath in conditionPathList:
      peakLibrary = spectConnect(conditionPath,options)
      conditionName = os.path.join(resultsPath,os.path.split(conditionPath)[1])
      mslFile = open(conditionName + '.MSL', 'w')
        # now named for the condition!
      peakLibrary.assignSizeRank()
      peakLibrary.outputMsl(mslFile)
      mslFile.close()
      if libraryFlag:
        peakLibrary.compareToMsl(mslPeakLibrary,options)
        mslFile = open(conditionName + '-translated.MSL', 'w')
        peakLibrary.outputMsl(mslFile)
        mslFile.close()
      totalLib.add(peakLibrary)

    if len(conditionPathList) > 1:
      minRA = 1000
      for library in totalLib.values():
        if library.minRA < minRA:
          minRA = library.minRA
      totalLib.minRA = minRA
      superLibrary = compareAllLibraries(totalLib, os.path.abspath(basePath), options)
      basePathName = os.path.split(os.path.abspath(basePath))[1]
      superLibraryFileName = os.path.join(basePath, os.path.join('results', \
        basePathName + '.MSL'))
      superLibraryFile = open(superLibraryFileName, 'w')
      superLibrary.outputMsl(superLibraryFile)
      superLibraryFile.close()
      conditionDictionaryFileName = os.path.join(basePath, \
        os.path.join('results','conditionList'))
      condDictFile = open(conditionDictionaryFileName, 'w')
      i = 1
      for conditionPath in conditionPathList:
        condDictFile.write(str(i) + ': ' + os.path.split(conditionPath)[1] + '\n')
        i += 1
      condDictFile.close()
      detailsFileName = os.path.join(basePath, os.path.join('results','details'))
      detailsFile = open(detailsFileName, 'w')
      superLibrary.outputDetailsFile(detailsFile)
      detailsFile.close()
      if options.alsoMatrix:
        generateMatrix(superLibrary, options)
    else:
      library = totalLib.values()[0]
      detailsFileName = os.path.join(basePath, os.path.join('results','details'))
      detailsFile = open(detailsFileName, 'w')
      library.outputDetailsFile(detailsFile)
      detailsFile.close()
      if options.alsoMatrix:
        generateMatrix(library, options)
      
def generateMatrix(mslPeakLibrary, options):
  for conditionPath in getConditionPathList(options.basePath):
    sampleLibrary = PeakLibraryLibrary(conditionPath)
    sampleLibrary.define()
    for eluPeakLibrary in sampleLibrary.values():
      mslPeakLibrary.conditionPath = conditionPath
      mslPeakLibrary.compareToElu(eluPeakLibrary,options)
  resultsPath = os.path.join(options.basePath, 'results')
  RAmatrixPath = os.path.join(resultsPath, 'RAmatrix.csv')
  ISmatrixPath = os.path.join(resultsPath, 'ISmatrix.csv')
  AMmatrixPath = os.path.join(resultsPath, 'AMmatrix.csv')
  mslPeakLibrary.outputMatrixDict(open(RAmatrixPath, 'w'),'RA')
  mslPeakLibrary.outputMatrixDict(open(ISmatrixPath, 'w'),'IS')
  mslPeakLibrary.outputMatrixDict(open(AMmatrixPath, 'w'),'AM')

def generatePathCombinations(pathList):
  combList = []
  for i in range(len(pathList)):
    for j in range(len(pathList[i+1:])):
      combList.append([pathList[i],pathList[i+1+j]])
  return combList

def getConditionPathList(basePath):
  conditionPathList = []
  for conditionPath in os.walk(basePath).next()[1]:
    conditionPathList.append(os.path.abspath(
      os.path.join(basePath,conditionPath)))
  conditionPathList.sort()
  return conditionPathList

def spectConnect(conditionPath,options):
  try:
    sampleLibrary = pickle.load(open(
      os.path.join(conditionPath, 'sampleLibrary.pickle')))
    patternLibrary = pickle.load(open(
      os.path.join(conditionPath, 'patternLibrary.pickle')))
  except:
    sampleLibrary = PeakLibraryLibrary(conditionPath)
    sampleLibrary.define()
    sampleLibrary.eliminateInternalMatches(options)
    gemoda = Gemoda(sampleLibrary,options)
    patternLibrary = gemoda.getPatternLibrary()
# Random debugging stuff below.
#   os.system('cp ' + os.path.join(conditionPath, 'blank.fa') + ' ' + \
#      os.path.join(conditionPath,'storage.fa'))
#    os.system('cp ' + os.path.join(conditionPath, 'blank.fa.go') + ' ' + \
#      os.path.join(conditionPath,'storage.fa.go'))
# Temp file storage to speed up redundant operations... made into a command-
#  line option, more for development than anything else.
    if options.keepTemps:
      sampPickle = open(os.path.join(conditionPath, 'sampleLibrary.pickle'), 'w')
      pattPickle = open(os.path.join(conditionPath, 'patternLibrary.pickle'), 'w')
      pickle.dump(sampleLibrary, sampPickle)
      pickle.dump(patternLibrary, pattPickle)
      sampPickle.close
      pattPickle.close

  peakLibrary = patternLibrary.getPeakLibrary()
  peakLibrary.eliminateInternalMatches(options) # this should not be necessary
  return peakLibrary
  
if __name__ == '__main__':
  main()

