
#include "realCompare.h"

/***********************************************************************
 * Calculates the outer product of the coordinates of each location in the
 *  two windows and sums them together for the correlation matrix.
 * Input: two gsl_matrices with the coordinates of <windowlength> alpha-
 *  carbons.
 * Output: a 3x3 gsl_matrix with the sum of the outer products.
 * **********************************************************************/
gsl_matrix *
sumOuterProd(gsl_matrix *a, gsl_matrix *b){
	int i,j,k;
	double newVal;
	gsl_matrix *h;
	// Make sure the matrices are the same size
	if(a->size1 != b->size1 || a->size2 != b->size2){
		fprintf(stderr, "Bad matrices for correlation matrix!");
		fflush(stderr);
		exit(0);
	}
	// Create a matrix to store the result in.
	h = gsl_matrix_calloc(a->size2, a->size2);

	// Now sum each outer product into the result matrix.
	for( i=0 ; i<a->size1 ; i++){
		for( j=0 ; j<a->size2 ; j++){
			for( k=0 ; k<a->size2 ; k++){
				newVal = gsl_matrix_get(h,j,k);
				newVal += gsl_matrix_get(a,i,j) * gsl_matrix_get(b,i,k);
				gsl_matrix_set(h, j, k, newVal);
			}
		}
	}
	return h;
}


/************************************************************************
 * Calculates the centroid of the gsl_matrix and moves the structure to
 *  have its centroid be at the origin.
 * Input: a gsl_matrix with the coordinates of two protein windows.
 * Output: Integer EXIT_SUCCESS value.
 * **********************************************************************/

int moveToCentroid(gsl_matrix *m){
	int i;
	gsl_vector *v = gsl_vector_calloc (m->size2);
	gsl_vector_view t;

	for(i=0; i<m->size1 ; i++){
		//  t is a row slice of m
		t = gsl_matrix_row(m, i);

		//  v is the sum of all rows
		gsl_vector_add(v, &t.vector);
	}
	//  v is now the average x,y,z
	gsl_vector_scale(v,  1/(double)m->size1);

	//  subtract off v from each row of m
	for(i=0; i<m->size1 ; i++){
		t = gsl_matrix_row(m, i);
		gsl_vector_sub(&t.vector, v);
	}
	gsl_vector_free(v);
	return EXIT_SUCCESS;
}

/**************************************************************************
 * Finds the matrix rotation to produce the optimal RMSD for two protein 
 *  windows that have their centroids at the origin.
 * Input: two gsl_matrices with alpha-carbon windows that we are
 *  looking to rotate.
 * Output: integer EXIT_SUCCESS (and a changed matrix b to have the optimal
 *  rotation).
 * ***********************************************************************/
int
rotateMats(gsl_matrix *a, gsl_matrix *b){
	gsl_matrix *h = NULL;
	gsl_vector *S = NULL;
	gsl_matrix *V = NULL;
	gsl_matrix *R = NULL;
	gsl_vector *work = NULL;
	gsl_matrix *result = NULL;
	CBLAS_TRANSPOSE_t noT = CblasNoTrans;
	CBLAS_TRANSPOSE_t yesT = CblasTrans;

	// find the correlation matrix h
	h = sumOuterProd(a,b);

	S = gsl_vector_alloc(h->size2);
	V = gsl_matrix_alloc(h->size2, h->size2);
	work = gsl_vector_alloc(h->size2);

	//  svd		
	//  h = U S V^T
	//
	//  h is replaced with U on exit
	//
	gsl_linalg_SV_decomp(h, V, S, work);

	R = gsl_matrix_calloc(h->size1, h->size1);
	gsl_blas_dgemm(noT,yesT, 1.0, V, h, 1.0, R);


	//  the final result matrix
	result = gsl_matrix_calloc(b->size1, h->size2);

	gsl_blas_dgemm(noT, noT, 1.0, b, R, 0.0, result);

	gsl_matrix_swap(b, result);

	gsl_vector_free(S);
	gsl_vector_free(work);
	gsl_matrix_free(V);
	gsl_matrix_free(h);
	gsl_matrix_free(R);
	gsl_matrix_free(result);
	return EXIT_SUCCESS;
}

/***********************************************************************
 * Computes the rmsd between two matrices.
 * Input: two gsl_matrices with alpha-carbon traces.
 * Output: a double with the value of the RMSD.
 * **********************************************************************/
double gsl_matrix_rmsd(gsl_matrix *a, gsl_matrix *b){
	int i,j;
	double total_rmsd=0;
	gsl_matrix *c = gsl_matrix_alloc(a->size1, a->size2);

	//  Check for reasonable inputs
	if(a->size1 != b->size1 || a->size2 != b->size2){
		fprintf(stderr, "Bad matrices for correlation matrix!");
		fflush(stderr);
		exit(0);
	}
	gsl_matrix_memcpy(c, a);

	gsl_matrix_sub(c, b);

	gsl_matrix_mul_elements(c, c);

	// Find the sum of the square deviations.
	for (i=0 ; i<c->size1 ; i++){
		for (j=0 ; j<c->size2 ; j++){
			total_rmsd += gsl_matrix_get(c, i, j);
		}
	}

	// Now get the rmsd.
	total_rmsd = sqrt( total_rmsd / c->size1 );
	gsl_matrix_free(c);
	return total_rmsd;
}


//  Calculate the rmsd between two windows, with optional
//  translation and rotation.
double
rmsdCompare(rdh_t *data, int win1, int win2, int L, double *extraParams){
	int trans = 1;
	int rot = 1;
	int dim;
	double result=0;
	int seq1,pos1;
	int seq2,pos2;
	gsl_matrix_view view1;
	gsl_matrix_view view2;
	gsl_matrix *mat1;
	gsl_matrix *mat2;
	gsl_matrix *mat1copy;
	gsl_matrix *mat2copy;

	// The "rint" function is in math.h and rounds a number to the 
	//  nearest integer.  It raises an "inexact exception" if the
	//  number initially wasn't an integer.
	if (extraParams != NULL) {
		trans = rint(extraParams[0]);
		rot = rint(extraParams[1]);
	}
	dim = getRdhDim(data);

	//  Find out which seq,pos pairs these two
	//  windows correspond to
	getRdhIndexSeqPos(data, win1, &seq1, &pos1);
	getRdhIndexSeqPos(data, win2, &seq2, &pos2);

	//  Get a reference to a submatrix.  That is,
	//  'chop out' the window.
	view1 = gsl_matrix_submatrix(data->seq[seq1], pos1, 0, L, dim);
	view2 = gsl_matrix_submatrix(data->seq[seq2], pos2, 0, L, dim);

	//  This just makes it easier to handle the views
	mat1 = &view1.matrix;
	mat2 = &view2.matrix;

	//  Create copies of the windows, because our comparison
	//  will require altering the matrices
	mat1copy = gsl_matrix_alloc(mat1->size1, mat1->size2);
	mat2copy = gsl_matrix_alloc(mat2->size1, mat2->size2);


	gsl_matrix_memcpy(mat1copy, mat1);
	gsl_matrix_memcpy(mat2copy, mat2);



	/*
	printf("matrix1:\n");
	gsl_matrix_pretty_fprintf(stdout, mat1copy, "%f ");
	printf("\nmatrix2:\n");
	gsl_matrix_pretty_fprintf(stdout, mat2copy, "%f ");
	*/

	//  Are we going to do a translation?
	if(trans == 1){
		moveToCentroid(mat1copy);
		moveToCentroid(mat2copy);
	}
	//  Are we going to do a rotation?
	if(rot == 1){
		//  Rotate mat2copy to have a minimal
		//  rmsd with mat1copy
		rotateMats(mat1copy,mat2copy);
	}

	//  Compute the rmsd between mat2copy and mat2copy
	result = gsl_matrix_rmsd(mat1copy, mat2copy);

	gsl_matrix_free(mat1copy);
	gsl_matrix_free(mat2copy);
	return result;
}
double
generalMatchFactor(rdh_t *data, int win1, int win2, int L, double *extraParams){
	int i,j;
	double numerator=0.0;
	/*double denominator=0.0;*/
	double xsum;
	double ysum;
	double ldenom=0.0;
	double rdenom=0.0;
	int dim;
	int seq1,pos1;
	int seq2,pos2;
	gsl_matrix_view view1;
	gsl_matrix_view view2;
	gsl_matrix *mat1;
	gsl_matrix *mat2;

	dim = getRdhDim(data);

	//  Find out which seq,pos pairs these two
	//  windows correspond to
	getRdhIndexSeqPos(data, win1, &seq1, &pos1);
	getRdhIndexSeqPos(data, win2, &seq2, &pos2);

	//  Get a reference to a submatrix.  That is,
	//  'chop out' the window.
	view1 = gsl_matrix_submatrix(data->seq[seq1], pos1, 0, L, dim);
	view2 = gsl_matrix_submatrix(data->seq[seq2], pos2, 0, L, dim);

	//  Some error checking here would be nice!
	//  Did we get the matrices we wanted?

	//  This just makes it easier to handle the views
	mat1 = &view1.matrix;
	mat2 = &view2.matrix;

	//  Loop over each position
	for ( i=0 ; i<mat1->size1 ; i++){
		xsum=0.0;
		ysum=0.0;

		//  Loop over each dimension at each position
		for ( j=0 ; j<dim ; j++){
			xsum += gsl_matrix_get(mat1, i, j);
			ysum += gsl_matrix_get(mat2, i, j);
		}

		numerator += (i+1) * sqrt(xsum * ysum);
		ldenom += (i+1) * xsum;
		rdenom += (i+1) * ysum;
	}

	return pow(numerator, 2.0)/(ldenom*rdenom);
}
double
massSpecCompareWElut(rdh_t *data, int win1, int win2, int L, double *extraParams){
	int i,j;
	double numerator=0.0;
	/*double denominator=0.0;*/
	double xsum;
	double ysum;
	double cum;
	double ldenom=0.0;
	double rdenom=0.0;
	int dim;
	int seq1,pos1;
	int seq2,pos2;
	double weight = 2.0;
	gsl_matrix_view view1;
	gsl_matrix_view view2;
	gsl_matrix *mat1;
	gsl_matrix *mat2;
	double maxElut = -1;
	int noSelfCompare = 0;
	
	if (extraParams != NULL) {
		maxElut = extraParams[0];
		noSelfCompare = extraParams[1];
	}

	dim = getRdhDim(data);

	//  Find out which seq,pos pairs these two
	//  windows correspond to
	getRdhIndexSeqPos(data, win1, &seq1, &pos1);
	getRdhIndexSeqPos(data, win2, &seq2, &pos2);

	if (noSelfCompare == 1) {
		if (seq1 == seq2) {
			if (pos1 != pos2) {
				return(0);
			}
		}
	}

	//  Get a reference to a submatrix.  That is,
	//  'chop out' the window.
	view1 = gsl_matrix_submatrix(data->seq[seq1], pos1, 0, L, dim);
	view2 = gsl_matrix_submatrix(data->seq[seq2], pos2, 0, L, dim);

	//  Some error checking here would be nice!
	//  Did we get the matrices we wanted?

	//  This just makes it easier to handle the views
	mat1 = &view1.matrix;
	mat2 = &view2.matrix;
	cum = 1.0;

	//  Loop over each position
	for ( i=0 ; i<mat1->size1 ; i++){
		xsum=0.0;
		ysum=0.0;

		// First take the first dimension for elution time
		if (maxElut >= 0) {
			if(fabs(gsl_matrix_get(mat1,i,0)-
					gsl_matrix_get(mat2,i,0)) > maxElut) {
				cum = 0;
				break;
			}
		}
//		printf("\n");
//
		//  Loop over each subsequent dimension at each position
		for ( j=1 ; j<dim ; j++){
//			printf("mat1val=%lf,mat2val=%lf\n",gsl_matrix_get(mat1,i,j),
//					gsl_matrix_get(mat2,i,j));
			numerator += pow(j,weight) * sqrt(gsl_matrix_get(mat1, i, j)
						* gsl_matrix_get(mat2,i,j));
			ldenom += pow(j,weight) * gsl_matrix_get(mat1, i, j);
			rdenom += pow(j,weight) * gsl_matrix_get(mat2, i, j);
//			printf("numer=%lf,ldenom=%lf,rdenom=%lf\n",numerator,
//					ldenom,rdenom);
		}

		cum *= pow(numerator, 2.0)/(ldenom*rdenom);
	}
	return pow(cum,1.0/L);
}

double (*getCompFunc(int compFunc))(rdh_t*,int,int,int,double*) {
	double (*comparisonFunc) (rdh_t*,int,int,int,double*) = &rmsdCompare;
	
	switch (compFunc) {
		case 0:
			comparisonFunc = &rmsdCompare;
			break;
		case 1:
			comparisonFunc = &generalMatchFactor;
			break;
		case 2:
			comparisonFunc = &massSpecCompareWElut;
			break;
		default:
			comparisonFunc = &rmsdCompare;
			break;
	}

	return(comparisonFunc);
}

bitGraph_t *
realComparison(rdh_t *data, int L, double g, int compFunc, double* extraParams){
	int i,j;
	int seq1,pos1;
	int seq2,pos2;
	bitGraph_t *bg = NULL;
	double score;
	double (*comparisonFunc) (rdh_t*,int,int,int,double*) = &rmsdCompare;

	//  Initialize the rdh's index
	initRdhIndex(data, L, 1);

	//  Allocate a new bit graph
	bg = newBitGraph(data->indexSize);

	// Choose the comparison function, pass a reference to it
	
	comparisonFunc = getCompFunc(compFunc);
		
	for( i=0 ; i<data->indexSize ; i++){

		//  Skip seperators
		getRdhIndexSeqPos(data, i, &seq1, &pos1);
		if(seq1 == -1 || pos1 == -1){
			continue;
		}

		for( j=i ; j<data->indexSize ; j++){
			getRdhIndexSeqPos(data, j, &seq2, &pos2);
			if(seq2 == -1 || pos2 == -1){
				continue;
			}

			//  This is the comparison function
			score = comparisonFunc(data, i, j, L, extraParams);
//			printf("score (%2d,%2d) vs. (%2d, %2d) =\t%lf\n",seq1, pos1, seq2, pos2, score);
			if (compFunc == 0) {
				if(score <= g) {
					bitGraphSetTrueSym(bg,i,j);
				}
			} else if ((compFunc == 1) || (compFunc == 2)) {
				if(score >=g){
					//printf("**\n");
					bitGraphSetTrueSym(bg,i,j);
				}
			} else {
				fprintf(stderr,"Comparison function undefined in "
							"realComparison function,\n located in "
							"realCompare.c.  Exiting.\n\n");
				fflush(stderr);
				exit(0);
			}
			/*printf("\n");*/
		}
	}

	return bg;
}

